/*
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * espnet.c
 * Part of espnet
 * Copyright (C) 2005 Simon Newton
 */

// required for vasprintf
// patches welcomed to remove it :)
#define _GNU_SOURCE

#include "private.h"
#include <stdio.h>
#include <stdarg.h>

char *espnet_errstr = NULL ;

/*
 * libespnet error function
 *
 * This writes the error string to espnet_strerror, which can be accessed
 * using espnet_strerror() ;
 *
 * Is vasprintf available on OSX ?
 */
void espnet_error(const char *fmt, ...) {
	
	va_list ap;
	va_start(ap, fmt) ;

	free(espnet_errstr) ;
	vasprintf(&espnet_errstr, fmt , ap) ,
	va_end(ap) ;

}


/*
 * Fetches the espnet error string
 * 
 * @return the error string
 */
char *espnet_strerror() {
	return espnet_errstr ;
}

/** 
 * Creates a new EspNet node.
 * Takes a string containing the ip address to bind to, if the string is NULL 
 * it uses the first non loopback address
 *
 * @param ip the IP address to bind to
 * @param debug level of logging provided 0: none
 * @return an espnet_node
 */
espnet_node espnet_new(const char *ip, int debug) {
	node n ;
	int ret ;
	
	n = malloc(sizeof(espnet_node_t ) ) ;
	
	if(n == NULL) {
		espnet_error("malloc failure") ;
		return NULL ;
	}

	memset(n, 0x0, sizeof(espnet_node_t) ) ;

	// default parameters
	n->state.mode = ESPNET_STANDBY ;
	strncpy(n->state.name, "libespnet", ESPNET_NAME_LENGTH) ;
	n->state.verbose = debug ;
	n->state.type = ESPNET_NODE_TYPE_IO ;	// wtf is this supposed to be, I want a multi in/out node damnit!
	n->state.strtcd = 0x00;					// default start code
	n->state.ttl = 4;						// default ttl
	n->state.tos = 0;						// default tos
	n->state.uni = 0;						// lets default to 0
	n->state.option = ESPNET_REPLY_OPTION_CONFIGURED ;
	n->state.bhr = ESPNET_OUT_BEHAV_BCAST ;
	n->state.peer_ip.s_addr = 0 ;			// no peer for now
	
	ret = espnet_net_init(n, ip) ;

	if (ret) {
		free(n);
		return NULL;
	}
	
	return (espnet_node) n;
}

/**
 * Starts the EspNet node.
 * Binds to the network sockets
 *
 * @param vn the espnet_node
 * @return 0 on success, non 0 on failure
 *
 */
int espnet_start(espnet_node vn) {
	node n = (node) vn ;
	int ret ;
	
	return_if_null(vn);
	return_if_on(n);

	if ((ret = espnet_net_start(n)) ) {
		return ret ;
	}
	n->state.mode = ESPNET_ON ;

	return ESPNET_EOK ;
}


/**
 * Stops the EspNet node
 * Closes the network sockets held by the node
 *
 * @param vn the espnet_node
 * @return 0 on success, non 0 on failure
 */
int espnet_stop(espnet_node vn) {
	node n = (node) vn ;
	int ret ;
	
	return_if_null(vn);
	return_if_off(n);
	
	if ((ret = espnet_net_close(n) )) 
		return ret ;

	n->state.mode = ESPNET_STANDBY ;

	return ESPNET_EOK; 
}


/**
 * Free the memory associated with this node
 *
 * @param vn the espnet_node
 */
int espnet_destroy(espnet_node vn) {

	return_if_null(vn);
	
	free(vn) ;
	return ESPNET_EOK;
}


/**
 * Handle any received packets.
 * This function is the workhorse of libespnet. You have a couple of options:
 * 	- use espnet_get_sd() to retrieve the socket descriptors and select to
 * 		detect network activity. Then call espnet_read(node,0) 
 * 		when activity is detected.
 * 	- call espnet_read repeatedly from within a loop with an appropriate 
 * 		timeout
 *
 * @param vn the espnet_node
 * @param timeout the number of seconds to block for if nothing is pending
 * @return 0 on success, -1 on failure
 */
int espnet_read(espnet_node vn, int timeout) {
	node n = (node) vn ;	
	espnet_packet_t p ;
	int ret ;
	
	return_if_null(vn);
	return_if_off(n);

	while(1) {
		memset( &p.data, 0x0, sizeof(p.data) ) ;
			
		if( (ret = espnet_net_recv(n, &p, timeout) ) < 0)
			return ret ;
		
		// nothing to read
		if (ret == RECV_NO_DATA) 
			break ;
		
		// skip this packet (filtered)
		if (p.length == 0)
			continue ;

		if( espnet_get_type(&p) ) {
			// should we be returning errors here?
			espnet_handle(n, &p) ;
		} 
	}
	return ESPNET_EOK ;
}



/**
 * This is used to set handlers for sent/received espnet packets.
 * If you're using a stock standard node you more than likely don't want
 * to set these. See the espnet_set_dmx_callback and espnet_set_firmware_callback .
 * If you want to get down and dirty with espnet packets, you can set this 
 * read / manipulate packets as they arrive (or get sent)
 *
 * @param vn 		The espnet_node
 * @param handler	The handler to set
 * @param fh		A pointer to a function, set to NULL to turn off
 * 						The function should return 0, 
 * @param data		Data to be passed to the handler when its called
 *
 * @return 0 on sucess, -1 on failure
 *
 */
int espnet_set_handler(espnet_node vn, espnet_handler_name_t handler, int (*fh)(espnet_node n, void *pp, void * d), void *data) {
	node n = (node) vn ;
	callback_t *callback ;

	return_if_null(vn);

	switch( handler) {
		case ESPNET_RECV_HANDLER:
			callback = &n->callbacks.recv ;
			break;
		case ESPNET_SEND_HANDLER:
			callback = &n->callbacks.send ;
			break;
		case ESPNET_POLL_HANDLER:
			callback = &n->callbacks.poll ;
			break ;
		case ESPNET_REPLY_HANDLER:
			callback = &n->callbacks.reply ;
			break ;
		case ESPNET_ACK_HANDLER:
			callback = &n->callbacks.ack ;
			break ;
		case ESPNET_DMX_HANDLER:
			callback = &n->callbacks.dmx ;
			break ;
		default:
			espnet_error("%s : Invalid handler defined", __FUNCTION__ ) ;
			return ESPNET_EARG;
	}
	callback->fh = fh ;
	callback->data = data;
	
	return ESPNET_EOK;
}


/*
 * This is a special callback which is invoked when dmx data is received. 
 * Unlike libartnet we pass the data back here, an esp node can recv on any
 * of 256 channels, we don't want to maintain buffer for them.
 *
 * @param vn 		The espnet_node
 * @param fh		The callback to invoke (parameters passwd are the espnet_node, the port_id
 * 					that received the dmx, and some user data
 * @param data		Data to be passed to the handler when its called
 *
 */
int espnet_set_dmx_handler(espnet_node vn, int (*fh)(espnet_node n, uint8_t universe, int length, uint8_t *data,  void *d), void *data ) {
	node n = (node) vn ;
	
	return_if_null(vn);

	n->callbacks.dmx_c.fh = fh ;
	n->callbacks.dmx_c.data = data;

	return ESPNET_EOK;
}



/**
 *
 * @param vn the espnet_node
 * @param ip the ip address to send to, NULL will broadcast the poll
 * @param type, type of reply requested, ESPNET_POLL_REQUEST_FULL or ESPNET_POLL_REQUEST_ACK
 */
int espnet_send_poll(espnet_node vn, const char *ip,  espnet_poll_request type) {
	node n = (node) vn ;
	
	return_if_null(vn);
	return_if_off(n) ;
	
	return espnet_tx_poll(n, ip, type) ;

}


/*
 * Sends some dmx data
 * 
 * @param vn the espnet_node
 * @param universe universe to address it to
 * @param length length of data
 * @param data pointer to dmx data
 *
 */
int espnet_send_dmx(espnet_node vn, uint8_t universe, int16_t length, uint8_t *data ) {	
	node n = (node) vn ;

	return_if_null(vn);
	return_if_off(n) ;

	if( length < 1 || length > ESPNET_DMX_LENGTH) {
		espnet_error("%s : Length of dmx data out of bounds (should be between 1 and 512, was %i)", __FUNCTION__ , length) ;
		return ESPNET_EARG;
	}

	if( n->state.type == ESPNET_NODE_TYPE_SINGLE_OUT) {
		espnet_error("%s : attempt to send from out output node", __FUNCTION__) ;
		return ESPNET_ESTATE;
	}

	if( n->state.type == ESPNET_NODE_TYPE_SINGLE_IN) {
		// force the universe
		return espnet_tx_dmx(n, n->state.uni, length, data) ;
	}

	return espnet_tx_dmx(n, universe, length, data) ;
	
}



/**
 * Sets the name of the node. 
 * The string should be null terminated and a maxmium of 18 Characters will be used
 * 
 * @param vn the espnet_node
 * @param name the short name of the node.
 */
int espnet_set_name(espnet_node vn, const char *name) {
	node n = (node) vn ;

	return_if_null(vn);

	strncpy((char *) &n->state.name, name, ESPNET_NAME_LENGTH) ;
	return ESPNET_EOK;
	
}

/*
 * Set the type of node
 *
 * @param vn	the espnet mode
 * @param typr	the node type, of type espnet_node_type
 */
int espnet_set_type(espnet_node vn, espnet_node_type type) {
	node n = (node) vn ;
	
	return_if_null(vn);

	// should be checking here
	n->state.type = type ;

	return ESPNET_EOK;
}


/*
 * Set the universe of this node, this can (well eventually) be
 * changed via the network.
 *
 * Single input/output nodes will only send/respond to datagrams that match this number
 *
 * @param vn	the espnet mode
 * @param typr	the universe

 */
int espnet_set_uni(espnet_node vn, uint8_t uni) {
	node n = (node) vn ;
	
	return_if_null(vn);

	n->state.uni = uni ;

	return ESPNET_EOK;
}

		
/**
 * Returns the socket descriptor associated with this espnet_node.
 * libespnet now uses a single descriptor per node, bound to the wildcard
 * address
 *
 * @param vn the espnet_node
 * @return the socket descriptor, or < 0 on error
 */
int espnet_get_sd(espnet_node vn) {
	node n = (node) vn ;
	
	return_if_null(vn);
	return_if_off(n);

	return n->sd ;
}


