/*
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * transmit.c
 * Functions to handle sending datagrams
 * Copyright (C) 2005 Simon Newton
 */

#include "private.h"

/*
 * Send an esp poll
 *
 * @param n 	the esp node
 * @param ip 	the ip address to send to (as a string)
 * @param type	type of request, of type espnet_poll_request
 */
int espnet_tx_poll(node n, const char *ip, espnet_poll_request type) {
	espnet_packet_t p ;

	if (n->state.mode != ESPNET_ON)
		return ESPNET_EACTION ;

	// check if caller specified an ip or wanted bcase
	if( ip != NULL) {
		if( ! inet_aton(ip, &p.to) ) {
			espnet_error("Ip conversion from %s failed", ip) ;
			return ESPNET_EARG ;
		}
	} else {
		p.to.s_addr = n->state.bcast_addr.s_addr ;
	}
	p.length = sizeof(espnet_poll_t) ;
	
	// contents
	p.data.poll.head = htonl(ESPNET_POLL) ;
	p.data.poll.type =  type ;

	return espnet_net_send(n, &p) ;

}


/*
 * Send an esp ack
 *
 * @param n			the esp node
 * @param dst		the address to send to
 * @param status	the status, of type espnet_ack_status
 * @param crc		the crc code
 *
 * ack's should be unicast (I think - protocol spec is vague on this)
 *
 */
int espnet_tx_ack(node n, SI dst, uint8_t status, uint8_t crc) {
	espnet_packet_t p ;

	if (n->state.mode != ESPNET_ON)
		return -1 ;

	p.to.s_addr = dst.s_addr ;
	p.length = sizeof(espnet_ack_t) ;

	// contents
	p.data.ack.head = htonl(ESPNET_ACK) ;
	p.data.ack.status = status;
	p.data.ack.crc = crc ;
	
	return espnet_net_send(n, &p) ;

}



/*
 * Send a poll reply
 *
 * @param n 	the node
 * @param dst	the dst to send to
 *
 */
int espnet_tx_poll_reply(node n, SI dst) {
	espnet_packet_t p ;

	memset(&p, 0x00, sizeof(p)) ;

	p.to.s_addr = dst.s_addr ;
	p.length = sizeof(espnet_poll_reply_t) ;

	// contents
	p.data.reply.head = htonl(ESPNET_REPLY) ;
	memcpy(	p.data.reply.mac, n->state.hw_addr, ESPNET_MAC_LENGTH)  ;
	p.data.reply.type = htons(n->state.type) ;
	p.data.reply.version = FIRMWARE_VER ;
	p.data.reply.sw = 0x00;
	memcpy(p.data.reply.name, n->state.name, ESPNET_NAME_LENGTH) ;
	p.data.reply.option = n->state.option ;
	p.data.reply.tos = n->state.tos;
	p.data.reply.ttl = n->state.ttl ;

	p.data.reply.config.listen = n->state.bhr ;
	memcpy(&p.data.reply.config.ip , &n->state.peer_ip.s_addr, 4) ;
	p.data.reply.config.uni = n->state.uni ;

	return espnet_net_send(n, &p) ; 
}


/*
 * Send a dmx datagram
 *
 * @param n			the esp node
 * @param universe	universe to address to
 * @param length	length of data
 * @param data		pointer to the data
 */
int espnet_tx_dmx(node n, uint8_t universe, int length, uint8_t *data) {
	espnet_packet_t p ;
		
	// should look at the bhv and adjust address here
	
	p.to.s_addr = n->state.bcast_addr.s_addr ;
	p.length = sizeof(espnet_data_t) - (ESPNET_DMX_LENGTH - length);
	
	// contents
	p.data.dmx.head = htonl(ESPNET_DMX) ;
	p.data.dmx.universe =  universe;
	p.data.dmx.start = n->state.strtcd ;
	
	if(n->state.bhr & ESPNET_IN_BEHAV_RLE) {
		// do rle and adjust p.length
		p.data.dmx.type = ESPNET_DMX_TYPE_RLE ;

	} else {
		p.data.dmx.type = ESPNET_DMX_TYPE_DATA ;
		p.data.dmx.size = htons(length) ;
		memcpy(&p.data.dmx.data, data, length) ; 
	}
	return espnet_net_send(n, &p) ;

}
